class UsageError(Exception):
    """
    ReArq usage error
    """


class ConfigurationError(Exception):
    """
    ReArq configuration error
    """


class SerializationError(Exception):
    """
    Serialization error
    """


class DeserializationError(Exception):
    """
    Deserialization error
    """
